"""Azure cli adaptor"""

# pylint: disable=import-outside-toplevel
import functools
import threading

azure = None
_session_creation_lock = threading.RLock()


def import_package(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global azure
        if azure is None:
            try:
                import azure as _azure  # type: ignore
                azure = _azure
            except ImportError:
                raise ImportError('Fail to import dependencies for Azure.'
                                  'Try pip install "skypilot[azure]"') from None
        return func(*args, **kwargs)

    return wrapper


@import_package
def get_subscription_id() -> str:
    """Get the default subscription id."""
    from azure.common import credentials
    return credentials.get_cli_profile().get_subscription_id()


@import_package
def get_current_account_user() -> str:
    """Get the default account user."""
    from azure.common import credentials
    return credentials.get_cli_profile().get_current_account_user()


@import_package
def http_error_exception():
    """HttpError exception."""
    from azure.core import exceptions
    return exceptions.HttpResponseError


@functools.lru_cache()
@import_package
def get_client(name: str, subscription_id: str):
    # Sky only supports Azure CLI credential for now.
    # Increase the timeout to fix the Azure get-access-token timeout issue.
    # Tracked in
    # https://github.com/Azure/azure-cli/issues/20404#issuecomment-1249575110
    from azure.identity import AzureCliCredential
    from azure.mgmt.network import NetworkManagementClient
    from azure.mgmt.resource import ResourceManagementClient
    with _session_creation_lock:
        credential = AzureCliCredential(process_timeout=30)
        if name == 'compute':
            from azure.mgmt.compute import ComputeManagementClient
            return ComputeManagementClient(credential, subscription_id)
        elif name == 'network':
            return NetworkManagementClient(credential, subscription_id)
        elif name == 'resource':
            return ResourceManagementClient(credential, subscription_id)
        else:
            raise ValueError(f'Client not supported: "{name}"')


@import_package
def create_security_rule(**kwargs):
    from azure.mgmt.network.models import SecurityRule
    return SecurityRule(**kwargs)



"""Rich status spinner utils."""
import contextlib
import threading
from typing import Union

import rich.console as rich_console

console = rich_console.Console()
_status = None

_logging_lock = threading.RLock()


class _NoOpConsoleStatus:
    """An empty class for multi-threaded console.status."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def update(self, text):
        pass

    def stop(self):
        pass

    def start(self):
        pass


def safe_status(msg: str) -> Union['rich_console.Status', _NoOpConsoleStatus]:
    """A wrapper for multi-threaded console.status."""
    from sky import sky_logging  # pylint: disable=import-outside-toplevel
    if (threading.current_thread() is threading.main_thread() and
            not sky_logging.is_silent()):
        global _status
        if _status is None:
            _status = console.status(msg)
        _status.update(msg)
        return _status
    return _NoOpConsoleStatus()


def force_update_status(msg: str):
    """Update the status message even if sky_logging.is_silent() is true."""
    if (threading.current_thread() is threading.main_thread() and
            _status is not None):
        _status.update(msg)


@contextlib.contextmanager
def safe_logger():
    logged = False
    with _logging_lock:
        if _status is not None and _status._live.is_started:  # pylint: disable=protected-access
            _status.stop()
            yield
            logged = True
            _status.start()
    if not logged:
        yield



"""Constants used for SkyServe."""

CONTROLLER_TEMPLATE = 'sky-serve-controller.yaml.j2'

SKYSERVE_METADATA_DIR = '~/.sky/serve'

# The filelock for selecting service ports on controller VM when starting a
# service. We need to have a filelock to avoid port collision when starting
# multiple services at the same time.
PORT_SELECTION_FILE_LOCK_PATH = f'{SKYSERVE_METADATA_DIR}/port_selection.lock'

# Signal file path for controller to handle signals.
SIGNAL_FILE_PATH = '/tmp/sky_serve_controller_signal_{}'

# Time to wait in seconds for service to register on the controller.
SERVICE_REGISTER_TIMEOUT_SECONDS = 60

# The time interval in seconds for load balancer to sync with controller. Every
# time the load balancer syncs with controller, it will update all available
# replica ips for each service, also send the number of requests in last query
# interval.
LB_CONTROLLER_SYNC_INTERVAL_SECONDS = 20

# Interval in seconds to probe replica endpoint.
ENDPOINT_PROBE_INTERVAL_SECONDS = 10

# The default timeout in seconds for a readiness probe request. We set the
# timeout to 15s since using actual generation in LLM services as readiness
# probe is very time-consuming (33B, 70B, ...).
# TODO(tian): Expose this option to users in yaml file.
READINESS_PROBE_TIMEOUT_SECONDS = 15

# Autoscaler window size in seconds for query per second. We calculate qps by
# divide the number of queries in last window size by this window size.
AUTOSCALER_QPS_WINDOW_SIZE_SECONDS = 60
# Autoscaler scale decision interval in seconds.
# We will try to scale up/down every `decision_interval`.
AUTOSCALER_DEFAULT_DECISION_INTERVAL_SECONDS = 20
# Autoscaler no replica decision interval in seconds.
AUTOSCALER_NO_REPLICA_DECISION_INTERVAL_SECONDS = 5
# Autoscaler default upscale delays in seconds.
# We will upscale only if the target number of instances
# is larger than the current launched instances for delay amount of time.
AUTOSCALER_DEFAULT_UPSCALE_DELAY_SECONDS = 300
# Autoscaler default downscale delays in seconds.
# We will downscale only if the target number of instances
# is smaller than the current launched instances for delay amount of time.
AUTOSCALER_DEFAULT_DOWNSCALE_DELAY_SECONDS = 1200
# The default controller resources. We need 200 GB disk space to enable using
# Azure as controller, since its default image size is 150 GB.
# TODO(tian): We might need to be careful that service logs can take a lot of
# disk space. Maybe we could use a larger disk size, migrate to cloud storage or
# do some log rotation.
CONTROLLER_RESOURCES = {'cpus': '4+', 'disk_size': 200}

# Due to the CPU/memory usage of the controller process launched with sky job (
# use ray job under the hood), we need to reserve some CPU/memory for each serve
# controller process.
# Serve: A default controller with 4 vCPU and 16 GB memory can run up to 16
# services.
CONTROLLER_MEMORY_USAGE_GB = 1.0

# A period of time to initialize your service. Any readiness probe failures
# during this period will be ignored.
DEFAULT_INITIAL_DELAY_SECONDS = 1200
DEFAULT_MIN_REPLICAS = 1

# Default port range start for controller and load balancer. Ports will be
# automatically generated from this start port.
CONTROLLER_PORT_START = 20001
LOAD_BALANCER_PORT_START = 30001
LOAD_BALANCER_PORT_RANGE = '30001-30100'

# Initial version of service.
INITIAL_VERSION = 1

# Replica ID environment variable name that can be accessed on the replica.
REPLICA_ID_ENV_VAR = 'SKYPILOT_SERVE_REPLICA_ID'

# The version of the lib files that serve use. Whenever there is an API
# change for the serve_utils.ServeCodeGen, we need to bump this version, so that
# the user can be notified to update their SkyPilot serve version on the remote
# cluster.
SERVE_VERSION = 1



"""Clouds need to be registered in CLOUD_REGISTRY to be discovered"""

import typing
from typing import Optional, Type

from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky.clouds import cloud


class _CloudRegistry(dict):
    """Registry of clouds."""

    def from_str(self, name: Optional[str]) -> Optional['cloud.Cloud']:
        if name is None:
            return None
        if name.lower() not in self:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Cloud {name!r} is not a valid cloud among '
                                 f'{list(self.keys())}')
        return self.get(name.lower())

    def register(self, cloud_cls: Type['cloud.Cloud']) -> Type['cloud.Cloud']:
        name = cloud_cls.__name__.lower()
        assert name not in self, f'{name} already registered'
        self[name] = cloud_cls()
        return cloud_cls


CLOUD_REGISTRY: _CloudRegistry = _CloudRegistry()



"""Credential checks: check cloud credentials and enable clouds."""
from typing import Dict, Iterable, Optional, Tuple

import click
import rich

from sky import clouds
from sky import global_user_state
from sky.adaptors import cloudflare


# TODO(zhwu): add check for a single cloud to improve performance
def check(quiet: bool = False, verbose: bool = False) -> None:
    echo = (lambda *_args, **_kwargs: None) if quiet else click.echo
    echo('Checking credentials to enable clouds for SkyPilot.')

    enabled_clouds = []

    def check_one_cloud(cloud_tuple: Tuple[str, clouds.Cloud]) -> None:
        cloud_repr, cloud = cloud_tuple
        echo(f'  Checking {cloud_repr}...', nl=False)
        ok, reason = cloud.check_credentials()
        echo('\r', nl=False)
        status_msg = 'enabled' if ok else 'disabled'
        styles = {'fg': 'green', 'bold': False} if ok else {'dim': True}
        echo('  ' + click.style(f'{cloud_repr}: {status_msg}', **styles) +
             ' ' * 30)
        if ok:
            enabled_clouds.append(cloud_repr)
            if verbose and cloud is not cloudflare:
                activated_account = cloud.get_current_user_identity_str()
                if activated_account is not None:
                    echo(f'    Activated account: {activated_account}')
            if reason is not None:
                echo(f'    Hint: {reason}')
        else:
            echo(f'    Reason: {reason}')

    clouds_to_check = [
        (repr(cloud), cloud) for cloud in clouds.CLOUD_REGISTRY.values()
    ]
    clouds_to_check.append(('Cloudflare, for R2 object store', cloudflare))

    for cloud_tuple in sorted(clouds_to_check):
        check_one_cloud(cloud_tuple)

    # Cloudflare is not a real cloud in clouds.CLOUD_REGISTRY, and should not be
    # inserted into the DB (otherwise `sky launch` and other code would error
    # out when it's trying to look it up in the registry).
    enabled_clouds = [
        cloud for cloud in enabled_clouds if not cloud.startswith('Cloudflare')
    ]
    global_user_state.set_enabled_clouds(enabled_clouds)

    if len(enabled_clouds) == 0:
        click.echo(
            click.style(
                'No cloud is enabled. SkyPilot will not be able to run any '
                'task. Run `sky check` for more info.',
                fg='red',
                bold=True))
        raise SystemExit()
    else:
        echo(
            click.style(
                '\nTo enable a cloud, follow the hints above and rerun: ',
                dim=True) + click.style('sky check', bold=True) + '\n' +
            click.style(
                'If any problems remain, refer to detailed docs at: '
                'https://skypilot.readthedocs.io/en/latest/getting-started/installation.html',  # pylint: disable=line-too-long
                dim=True))

        # Pretty print for UX.
        if not quiet:
            enabled_clouds_str = '\n  :heavy_check_mark: '.join(
                [''] + sorted(enabled_clouds))
            rich.print('\n[green]:tada: Enabled clouds :tada:'
                       f'{enabled_clouds_str}[/green]')


def get_cloud_credential_file_mounts(
        excluded_clouds: Optional[Iterable[clouds.Cloud]]) -> Dict[str, str]:
    """Returns the files necessary to access all enabled clouds.

    Returns a dictionary that will be added to a task's file mounts
    and a list of patterns that will be excluded (used as rsync_exclude).
    """
    enabled_clouds = global_user_state.get_enabled_clouds()
    file_mounts = {}
    for cloud in enabled_clouds:
        if (excluded_clouds is not None and
                clouds.cloud_in_list(cloud, excluded_clouds)):
            continue
        cloud_file_mounts = cloud.get_credential_file_mounts()
        file_mounts.update(cloud_file_mounts)
    # Currently, get_enabled_clouds() does not support r2
    # as only clouds with computing instances are marked
    # as enabled by skypilot. This will be removed when
    # cloudflare/r2 is added as a 'cloud'.
    r2_is_enabled, _ = cloudflare.check_credentials()
    if r2_is_enabled:
        r2_credential_mounts = cloudflare.get_credential_file_mounts()
        file_mounts.update(r2_credential_mounts)
    return file_mounts



"""A script that generates the Fluidstack catalog.

Usage:
    python fetch_fluidstack_cloud.py
"""

import copy
import csv
import json
import os
from typing import List

import requests

ENDPOINT = 'https://api.fluidstack.io/v1/plans'
DEFAULT_FLUIDSTACK_API_KEY_PATH = os.path.expanduser('~/.fluidstack/api_key')
DEFAULT_FLUIDSTACK_API_TOKEN_PATH = os.path.expanduser(
    '~/.fluidstack/api_token')

GPU_MAP = {
    'H100_PCIE_80GB': 'H100',
    'A100_SXM4_80GB': 'A100-80GB',
    'A100_PCIE_80GB': 'A100-80GB',
    'A100_SXM4_40GB': 'A100',
    'A100_PCIE_40GB': 'A100',
    'Tesla_V100_SXM2_16GB': 'V100',
    'Tesla_V100_PCIE_16GB': 'V100',
    'A10_PCIE_24GB': 'A10',
    'A30_24GB': 'A30',
    'A40_48GB': 'A40',
    'RTX_A6000_48GB': 'RTXA6000',
    'RTX_A5000_24GB': 'RTXA5000',
    'RTX_A4000_16GB': 'RTXA4000',
    'Quadro_RTX_5000_16GB': 'RTX5000',
    'Quadro_RTX_4000_8GB': 'RTX4000',
    'L40_48GB': 'L40',
    'Quadro_RTX_6000_16GB': 'RTX6000',
    'T4_16GB': 'T4',
    'RTX_3090_24GB': 'RTX3090',
    'RTX_3080_10GB': 'RTX3080',
}

CUSTOM_PLANS_CONFIG = [
    dict(gpu_count=1, cpu_count=8, nvme_storage=750, ram=64),
    dict(gpu_count=2, cpu_count=16, nvme_storage=1024, ram=128),
    #dict(gpu_count=4, cpu_count=32, nvme_storage=1200, ram=160),
    #dict(gpu_count=8, cpu_count=64, nvme_storage=1500, ram=200)
]


def get_regions(plans: List) -> dict:
    """Return a list of regions where the plan is available."""
    regions = {}
    for plan in plans:
        for region in plan.get('regions', []):
            regions[region['id']] = region['id']
    return regions


def plans_from_custom_plan(plan: dict) -> List[dict]:
    prices = dict(cpu=plan['price']['cpu']['hourly'],
                  ram=plan['price']['ram']['hourly'],
                  storage=plan['price']['storage']['hourly'],
                  gpu=plan['price']['gpu']['hourly'])
    new_plans = []
    for i, config in enumerate(CUSTOM_PLANS_CONFIG):
        new_plan = copy.deepcopy(plan)
        price = (prices['cpu'] *
                 config['cpu_count']) + (prices['ram'] * config['ram']) + (
                     prices['storage'] * config['nvme_storage']) + (
                         prices['gpu'] * config['gpu_count'])
        new_plan['price']['hourly'] = price / config['gpu_count']
        new_plan['configuration']['core_count'] = config['cpu_count']
        new_plan['configuration']['ram'] = config['ram']
        new_plan['configuration']['gpu_count'] = config['gpu_count']
        new_plan['configuration']['nvme_storage'] = config['nvme_storage']
        new_plan['plan_id'] = f'custom:{i}:{plan["plan_id"]}'
        new_plans.append(new_plan)
    return new_plans


def create_catalog(output_dir: str) -> None:
    response = requests.get(ENDPOINT)
    plans = response.json()
    custom_plans = [
        plan for plan in plans if plan['minimum_commitment'] == 'hourly' and
        plan['type'] in ['custom'] and plan['gpu_type'] != 'NO GPU'
    ]
    #plans = [plan for plan in plans if len(plan['regions']) > 0]
    plans = [
        plan for plan in plans if plan['minimum_commitment'] == 'hourly' and
        plan['type'] in ['preconfigured'] and
        plan['gpu_type'] not in ['NO GPU', 'RTX_3080_10GB', 'RTX_3090_24GB']
    ]

    plans = plans + [
        plan for plan in custom_plans for plan in plans_from_custom_plan(plan)
    ]
    with open(os.path.join(output_dir, 'vms.csv'), mode='w',
              encoding='utf-8') as f:
        writer = csv.writer(f, delimiter=',', quotechar='"')
        writer.writerow([
            'InstanceType',
            'AcceleratorName',
            'AcceleratorCount',
            'vCPUs',
            'MemoryGiB',
            'Price',
            'Region',
            'GpuInfo',
            'SpotPrice',
        ])
        for plan in plans:
            try:
                gpu = GPU_MAP[plan['gpu_type']]
            except KeyError:
                print(f'Could not map {plan["gpu_type"]}')
                continue
            gpu_memory = int(
                str(plan['configuration']['gpu_memory']).replace('GB',
                                                                 '')) * 1024
            gpu_cnt = int(plan['configuration']['gpu_count'])
            vcpus = float(plan['configuration']['core_count'])
            mem = float(plan['configuration']['ram'])
            price = float(plan['price']['hourly']) * gpu_cnt
            gpuinfo = {
                'Gpus': [{
                    'Name': gpu,
                    'Manufacturer': 'NVIDIA',
                    'Count': gpu_cnt,
                    'MemoryInfo': {
                        'SizeInMiB': int(gpu_memory)
                    },
                }],
                'TotalGpuMemoryInMiB': int(gpu_memory * gpu_cnt),
            }
            gpuinfo = json.dumps(gpuinfo).replace('"', "'")  # pylint: disable=invalid-string-quote
            for r in plan.get('regions', []):
                if r['id'] == 'india_2':
                    continue
                writer.writerow([
                    plan['plan_id'],
                    gpu,
                    gpu_cnt,
                    vcpus,
                    mem,
                    price,
                    r['id'],
                    gpuinfo,
                    '',
                ])


if __name__ == '__main__':

    os.makedirs('fluidstack', exist_ok=True)
    create_catalog('fluidstack')
    print('Fluidstack catalog saved to {}/vms.csv'.format('fluidstack'))



"""Oracle OCI cloud adaptor"""

import functools
import os

CONFIG_PATH = '~/.oci/config'
ENV_VAR_OCI_CONFIG = 'OCI_CONFIG'

oci = None


def import_package(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global oci
        if oci is None:
            try:
                # pylint: disable=import-outside-toplevel
                import oci as _oci
                oci = _oci
            except ImportError:
                raise ImportError('Fail to import dependencies for OCI.'
                                  'Try pip install "skypilot[oci]"') from None
        return func(*args, **kwargs)

    return wrapper


def get_config_file() -> str:
    conf_file_path = CONFIG_PATH
    config_path_via_env_var = os.environ.get(ENV_VAR_OCI_CONFIG)
    if config_path_via_env_var is not None:
        conf_file_path = config_path_via_env_var
    return conf_file_path


@import_package
def get_oci():
    return oci


@import_package
def get_oci_config(region=None, profile='DEFAULT'):
    conf_file_path = get_config_file()
    oci_config = oci.config.from_file(file_location=conf_file_path,
                                      profile_name=profile)
    if region is not None:
        oci_config['region'] = region
    return oci_config


@import_package
def get_core_client(region=None, profile='DEFAULT'):
    return oci.core.ComputeClient(get_oci_config(region, profile))


@import_package
def get_net_client(region=None, profile='DEFAULT'):
    return oci.core.VirtualNetworkClient(get_oci_config(region, profile))


@import_package
def get_search_client(region=None, profile='DEFAULT'):
    return oci.resource_search.ResourceSearchClient(
        get_oci_config(region, profile))


@import_package
def get_identity_client(region=None, profile='DEFAULT'):
    return oci.identity.IdentityClient(get_oci_config(region, profile))


@import_package
def service_exception():
    """OCI service exception."""
    # pylint: disable=import-outside-toplevel
    from oci.exceptions import ServiceError
    return ServiceError
